
local addon = LibStub("AceAddon-3.0"):NewAddon("ShuffleThoseIcons", "AceTimer-3.0", "AceEvent-3.0")
_G.ShuffleThoseIcons = addon

local common = {
	enable = {
		type = "toggle",
		name = "Enable",
		desc = "Enable or disable this icon shuffler.",
		order = 1,
	},
	name = {
		type = "input",
		name = "Name",
		desc = "Give this shuffler a unique name so you can easily find it later.",
		order = 2,
	},
	target = {
		type = "input",
		name = "Target",
		desc = "Name of the unit, aura or spell to act upon.",
		order = 3,
	},
	event = {
		type = "select",
		name = "Event",
		desc = "Select the event upon which to dish out the icons.",
		values = {
			SPELL_SUMMON = "Unit summoned",
			SPELL_AURA_APPLIED = "Aura applied",
			SPELL_CAST_START = "Spell casting",
		},
		style = "dropdown",
		order = 4,
	},
	timeout = {
		type = "range",
		name = "Timeout",
		desc = "How long in seconds the shuffler should keep trying to mark before resetting to the lowest index icon on the next time it is triggered. Select a timeout of 0 to reset instantly after the first icon is dealt.",
		min = 0,
		max = 30,
		step = 1,
		bigStep = 5,
		order = 5,
	},
	icons = {
		type = "multiselect",
		name = "Icons",
		desc = "Select all the icons that the shuffler can use. It will start with the lowest index one (Star) and go up to the highest (Skull).",
		values = {
			"Star",
			"Circle",
			"Diamond",
			"Triangle",
			"Moon",
			"Square",
			"Cross",
			"Skull",
		},
		order = 6,
		style = "dropdown",
		dialogControl = "Dropdown",
	},
	enemy = {
		type = "toggle",
		name = "Mark enemies",
		desc = "Toggles whether this shuffler is capable of marking enemies.",
		order = 7,
	},
	friendly = {
		type = "toggle",
		name = "Mark friendlies",
		desc = "Toggles whether this shuffler is capable of marking friendlies.",
		order = 8,
	},
	create = {
		type = "execute",
		name = "Create",
		desc = "Creates the specified shuffler.",
		order = 9,
	},
	delete = {
		type = "execute",
		name = "Delete",
		desc = "Deletes this shuffler.",
		order = 10,
	},
}

local defaults = {
	name = "",
	target = "",
	event = "SPELL_AURA_APPLIED",
	timeout = 5,
	icons = { true, true, true, true, true, true, true, true },
	enemy = true,
	friendly = true,
}
local newShuffler = {}
local function getDefaults(info, key, ...)
	local option = info[#info]
	local x = type(newShuffler[option]) == "nil" and defaults[option] or newShuffler[option]
	if type(x) == "table" then return x[key]
	else return x end
end
local function setNew(info, first, second)
	local option = info[#info]
	if type(second) ~= "nil" then -- It's the multiselect
		if not newShuffler[option] then
			newShuffler[option] = { true, true, true, true, true, true, true, true }
		end
		newShuffler[option][first] = second
	else
		newShuffler[option] = first
	end
end

local options = {
	type = "group",
	childGroups = "tree",
	hidden = function(info)
		local option = info[#info]
		local parent = info[#info - 1]
		if parent == "new" then
			if option == "enable" or option == "delete" then return true end
		elseif parent ~= "new" then
			if option == "name" or option == "create" then return true end
		end
	end,
	args = {
		desc = {
			type = "description",
			name = "Shuffle Those Icons lets you create semi-scripted triggers and handlers for automatic raid icon application.\n\n",
			order = 1,
		},
		new = {
			type = "group",
			name = "Create new",
			order = 2,
			args = common,
			set = setNew,
			get = getDefaults,
		},
	},
}
local function populateOptions()
	for k in pairs(options.args) do
		if k ~= "desc" and k ~= "new" then
			options.args[k] = nil
		end
	end
	for name, shuffler in pairs(addon.db.profile) do
		options.args[name] = {
			type = "group",
			name = name,
			order = 3,
			args = common,
			set = function(info, first, second)
				local option = info[#info]
				if type(second) ~= "nil" then -- It's the multiselect
					addon.db.profile[name][option][first] = second
				else
					addon.db.profile[name][option] = first
				end
			end,
			get = function(info, key)
				local option = info[#info]
				local x = addon.db.profile[name][option]
				if type(x) == "table" then return x[key]
				else return x end
			end,
		}
	end
end

local function deleteShuffler()
	print("oO")
end

local function createNewShuffler()
	-- Validate that everything required is set
	local valid = true
	if not newShuffler.name or newShuffler.name:trim():len() == 0 then valid = false
	elseif not newShuffler.target or newShuffler.target:trim():len() == 0 then valid = false
	end
	if not valid then
		print("You forgot to fill in the name or target of the shuffler.")
		return
	end
	if addon.db.profile[newShuffler.name] then
		print("A shuffler with this name exists already.")
		return
	end
	addon.db.profile[newShuffler.name] = {
		target = newShuffler.target,
		event = newShuffler.event or defaults.event,
		timeout = newShuffler.timeout or defaults.timeout,
		icons = {},
		enemy = type(newShuffler.enemy) == "nil" and true or newShuffler.enemy,
		friendly = type(newShuffler.friendly) == "nil" and true or newShuffler.friendly,
		enable = true,
	}
	local icons = newShuffler.icons or defaults.icons
	for i, v in next, icons do
		addon.db.profile[newShuffler.name].icons[i] = v
	end
	wipe(newShuffler)
	populateOptions()
end

common.create.func = createNewShuffler
common.delete.func = deleteShuffler

local findTargetByGUID = nil
do
	local t = {"target", "targettarget", "focus", "focustarget", "mouseover", "mouseovertarget"}
	local fmt = string.format
	for i = 1, 4 do t[#t+1] = fmt("boss%d", i) end
	for i = 1, 4 do t[#t+1] = fmt("party%dtarget", i) end
	for i = 1, 40 do t[#t+1] = fmt("raid%dtarget", i) end
	fmt = nil
	function findTargetByGUID(id)
		for i, unit in next, t do
			local unitId = UnitGUID(unit)
			if unitId and unitId == id then return unit end
		end
	end
end

local function findNextMarkIndex(shuffler)
	-- Find first unused mark index
	for i, index in ipairs(shuffler.icons) do
		if index and (not shuffler.lastMark or shuffler.lastMark < i) then
			shuffler.lastMark = i
			return i
		end
	end
	-- All mark indexes used, start over
	for i, index in ipairs(shuffler.icons) do
		if index then
			shuffler.lastMark = i
			return i
		end
	end
end

local repeatingScanTimer = nil
local targetsToFind = {}
local stopScanningForID = {}

local function repeatingScanner()
	local t = GetTime()
	for guid, icon in pairs(targetsToFind) do
		local id = findTargetByGUID(guid)
		if id then
			SetRaidTarget(id, icon)
			targetsToFind[guid] = nil
		else
			if t > stopScanningForID[guid] then
				targetsToFind[guid] = nil
				stopScanningForID[guid] = nil
			end
		end
	end
	if not next(targetsToFind) then
		addon:CancelTimer(repeatingScanTimer)
		repeatingScanTimer = nil
	end
end

local function markTarget(shuffler, guid, name, friendly, enemy)
	local icon = findNextMarkIndex(shuffler)
	if not icon then error("Crash and burn, sucker!") end
	if friendly then
		-- If it's a friendly target we can just mark by name.
		SetRaidTarget(name, icon)
	else
		-- It's an enemy target, we need to scan for it and find a valid unitId.
		local id = findTargetByGUID(guid)
		if id then
			SetRaidTarget(id, icon)
		else
			if not repeatingScanTimer then
				repeatingScanTimer = addon:ScheduleRepeatingTimer(repeatingScanner, 0.2)
			end
			targetsToFind[guid] = icon
			stopScanningForID[guid] = (GetTime() + v.timeout + 0.3)
		end
	end
	
end

function addon:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, sGUID, source, sFlags, dGUID, dName, dFlags, spellId, spellName, _, secSpellId, buffStack)
	if event ~= "SPELL_AURA_APPLIED" and event ~= "SPELL_SUMMON" and event ~= "SPELL_CAST_START" then return end
	for k, v in pairs(self.db.profile) do
		if event == v.event and (v.target == spellName or v.target == dName) then
			if v.target == dName and event == "SPELL_SUMMON" then
				local friend = CombatLog_Object_IsA(dFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)
				local enemy = CombatLog_Object_IsA(dFlags, COMBATLOG_OBJECT_REACTION_HOSTILE)
				if (v.enemy and enemy) or (v.friendly and friend) then
					markTarget(v, dGUID, dName, friend, enemy)
				end
			elseif v.target == spellName then
				if event == "SPELL_AURA_APPLIED" then
					local friend = CombatLog_Object_IsA(dFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)
					local enemy = CombatLog_Object_IsA(dFlags, COMBATLOG_OBJECT_REACTION_HOSTILE)
					if (v.enemy and enemy) or (v.friendly and friend) then
						markTarget(v, dGUID, dName, friend, enemy)
					end
				elseif event == "SPELL_CAST_START" then
					local friend = CombatLog_Object_IsA(sFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)
					local enemy = CombatLog_Object_IsA(sFlags, COMBATLOG_OBJECT_REACTION_HOSTILE)
					if (v.enemy and enemy) or (v.friendly and friend) then
						markTarget(v, sGUID, source, friend, enemy)
					end
				end
			end
		end
	end
end

-- Should use GetSpellInfo for the target spell names in the dbDefaults table,
-- just so the defaults work for all locales, but fuck me I am lazy.
local dbDefaults = {
	profile = {
		["ICC LK Valkyrs"] = {
			target = "Val'kyr Shadowguard",
			event = "SPELL_SUMMON",
			timeout = 10,
			icons = { true, true, true, false, false, false, false, false },
			enemy = true,
			friendly = false,
			enable = true,
		},
		["ICC Deathwhisper MC"] = {
			target = "Dominate Mind",
			event = "SPELL_AURA_APPLIED",
			timeout = 5,
			icons = { true, true, true, false, false, false, false, false },
			enemy = true,
			friendly = true,
			enable = true,
		},
		["ICC Sindragosa Tombs"] = {
			target = "Frost Beacon",
			event = "SPELL_AURA_APPLIED",
			timeout = 0,
			icons = { true, true, true, true, true, true, false, false },
			enemy = false,
			friendly = true,
			enable = true,
		},
		["ICC Festergut Spores"] = {
			target = "Gas Spore",
			event = "SPELL_AURA_APPLIED",
			timeout = 0,
			icons = { true, true, true, false, false, false, false, false },
			enemy = false,
			friendly = true,
			enable = true,
		},
		["ICC BQL Pact"] = {
			target = "Pact of the Darkfallen",
			event = "SPELL_AURA_APPLIED",
			timeout = 0,
			icons = { true, true, false, false, false, false, false, false },
			enemy = false,
			friendly = true,
			enable = true,
		},
		["ICC Council Target"] = {
			target = "Invocation of Blood",
			event = "SPELL_AURA_APPLIED",
			timeout = 10,
			icons = { false, false, false, false, false, false, false, true },
			enemy = true,
			friendly = false,
			enable = true,
		},
		["ICC Putricide Plague"] = {
			target = "Unbound Plague",
			event = "SPELL_AURA_APPLIED",
			timeout = 0,
			icons = { false, false, false, false, false, false, true, false },
			enemy = false,
			friendly = true,
			enable = true,
		},
		["ICC Rotface Infection"] = {
			target = "Mutated Infection",
			event = "SPELL_AURA_APPLIED",
			timeout = 0,
			icons = { false, false, false, false, false, false, false, true },
			enemy = false,
			friendly = true,
			enable = true,
		},
	},
}

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("STIDB", dbDefaults, true)

	for k, v in pairs(self.db.profile) do
		v.lastMark = nil
	end

	populateOptions()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("ShuffleThoseIcons", options)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ShuffleThoseIcons", "ShuffleThoseIcons")
end

function addon:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end


